#!/usr/bin/env python3

"""collection of Version objects used by crmngrlib"""


class Version:
    """Base class for version objects"""

    def __init__(self, version):
        """Initialize Version
        :argument version Version(-string) for this module.
        """
        self._version = version

    def __hash__(self):
        hash(self._version)

    @property
    def version(self):
        """Return Version(-string)"""
        return self._version

    @property
    def version_type(self):
        """String representation of version type"""
        return type(self).__name__


class Forge(Version):
    """Puppet Forge Version"""

    def __repr__(self):
        return "%s" % self._version

    @property
    def puppetfile(self):
        """Return version in suitable format for puppetfile"""
        return "'%s'" % self.version


class GitBranch(Version):
    """Git Branch"""

    def __repr__(self):
        return "branch:%s" % self._version

    @property
    def puppetfile(self):
        """Return version in suitable format for puppetfile"""
        return "  :branch => '%s'" % self.version


class GitCommit(Version):
    """Git Commit"""

    def __repr__(self):
        return "commit:%s" % self._version

    @property
    def puppetfile(self):
        """Return version in suitable format for puppetfile"""
        return "  :commit => '%s'" % self.version


class GitRef(Version):
    """Git Ref"""

    def __repr__(self):
        return "ref:%s" % self._version

    @property
    def puppetfile(self):
        """Return version in suitable format for puppetfile"""
        return "  :ref => '%s'" % self.version


class GitTag(Version):
    """ Git Tag"""

    def __repr__(self):
        return "tag:%s" % self._version

    @property
    def puppetfile(self):
        """Return version in suitable format for puppetfile"""
        return "  :tag => '%s'" % self.version




#!/usr/bin/env python3

"""collection of PuppetModule objects used by crmngrlib"""

from .versions import Version, Forge, GitCommit, GitBranch, GitTag, GitRef
from ..utils import cprint


class PuppetModule:
    """Base class for puppet modules"""

    def __init__(self, name):
        """Initialize puppet module
        :argument name Name of module
        """
        self._name = name
        self._info = {}
        self._branches = []
        self._version = None

    @property
    def info(self):
        """Get external module info"""
        return self._info

    @info.setter
    def info(self, value):
        """Set external module info"""
        self._info = value

    @property
    def name(self):
        """Name of this module"""
        return self._name

    @property
    def branches(self):
        """Branches that use this module"""
        return self._branches

    @property
    def version(self):
        """Return this modules version"""
        return self._version

    @version.setter
    def version(self, value):
        """Set version of this puppet module
        :argument value version (needs to be a subclass of Version)
        """
        if isinstance(value, Version):
            self._version = value
        else:
            if value is None:
                self._version = None
            else:
                raise AttributeError('Unsupported type %s for value' % type(value))

    def __hash__(self):
        return hash(self.__repr__())

    def __repr__(self):
        return "%s" % self.name

    def __eq__(self, other):
        return str(self) >= str(other) >= str(self)

    def __ne__(self, other):
        return str(self) < str(other) or str(other) < str(self)

    def __gt__(self, other):
        return str(other) < str(self)

    def __ge__(self, other):
        return str(self) >= str(other)

    def __le__(self, other):
        return str(other) >= str(self)

    def add_branch(self, branch):
        """Add a branch to this modules internal branch list
        :argument branch Name of branch to add
        """
        if branch not in self._branches:
            self._branches.append(branch)


class GitModule(PuppetModule):
    """Puppet mdoule hosted on git"""

    def __init__(self, name, url):
        """Initialize git module
        :argument name Name of module
        :argument url Repository URL of module
        """
        super().__init__(name)
        self._url = url

    @property
    def url(self):
        """URL to git repository for this puppet module"""
        return self._url

    def __repr__(self):
        """Return unique string representation"""
        representation = "%s:git:%s" % (self.name, self.url)
        if self.version:
            representation += ":%s" % self.version
        return representation

    @property
    def puppetfile(self):
        """Return puppetfile representation of module"""
        lines = [
            "mod '%s'," % self.name,
        ]
        git_line = "  :git => '%s'" % self.url
        if self.version:
            git_line += ","
        lines.append(git_line)

        if isinstance(self.version, Version):
            lines.append(self.version.puppetfile)
        return lines

    def print_version_information(self):
        """Print out version information"""
        cprint.magenta_bold('Version:', lpad=2)
        cprint.white('Git:', lpad=4, rpad=8, end='')
        cprint.white(self.url)
        if isinstance(self.version, GitBranch):
            cprint.blue_bold('Branch: ', lpad=16, end='')
            cprint.yellow_bold(self.version.version, end='')
        elif isinstance(self.version, GitCommit):
            cprint.red('Commit: ', lpad=16, end='')
            cprint.red_bold(self.version.version[:7], end='')
        elif isinstance(self.version, GitRef):
            cprint.red('Ref: ', lpad=16, end='')
            cprint.red_bold(self.version.version, end='')
        elif isinstance(self.version, GitTag):
            cprint.blue_bold('Tag: ', lpad=16, end='')
            if self.info.get('version', None) == self.version.version:
                cprint.green_bold(self.version.version, end='')
            else:
                cprint.yellow_bold(self.version.version, end='')
        else:
            cprint.red_bold('UNSPECIFIED', lpad=16, end='')
        cprint.white('[Latest: %s (%s)]' % (
            self.info.get('version', 'unknown'),
            self.info.get('date', 'unknown')
        ), lpad=1)


class ForgeModule(PuppetModule):
    """Puppet module hosted on forge"""

    def __init__(self, name, author):
        """Initialize forge module
        :argument name Name of module
        :argument author Author/Namespace of the module
        """
        super().__init__(name)
        self._author = author

    @property
    def author(self):
        """Puppet module author / namespace"""
        return self._author

    @property
    def forgename(self):
        """Return module name as used on forge"""
        return "%s/%s" % (self._author, self._name)

    def __repr__(self):
        """Return unique string representation"""
        representation = "%s:forge:%s" % (self.name, self.author)
        if self.version:
            representation += ":%s" % self.version
        return representation

    def print_version_information(self):
        """Print out version information"""
        cprint.magenta_bold('Version:', lpad=2)
        cprint.white('Forge:', lpad=4, rpad=6, end='')
        cprint.white(self.forgename, suffix=':')
        if not self.version:
            cprint.red_bold('UNSPECIFIED', lpad=16, end='')
        else:
            if self.info.get('version', None) == self.version.version:
                cprint.green_bold(self.version.version, lpad=16, end='')
            else:
                cprint.yellow_bold(self.version.version, lpad=16, end='')
        cprint.white(' [Latest: %s (%s)]' % (
            self.info.get('version', 'unknown'),
            self.info.get('date', 'unknown')
        ))

    @property
    def puppetfile(self):
        """Return puppetfile representation of module"""
        line = "mod '%s/%s'" % (self.author, self.name)
        if self.version:
            line += ", %s" % self.version.puppetfile
        return [line, ]



#!/usr/bin/env python3

"""git related functions for crmngrlib"""

import logging
import os
import shutil
import subprocess

from datetime import datetime

from .cache import JsonCache
from .utils import sha256

LOG = logging.getLogger(__name__)


class Git:
    """git object, holding cache and tmpdir information"""

    def __init__(self, tmpdir, cache=None, git_binary='git'):
        self._cache = cache
        self._controlbranches = []
        self._git_binary = git_binary
        self._tmpdir = tmpdir
        self._workdir = ""

    @property
    def controlbranches(self):
        """controlbranches in this git repository"""
        return self._controlbranches

    @controlbranches.setter
    def controlbranches(self, value):
        """controlbranches in this git repository (setter)"""
        self._controlbranches = value

    def cmd(self, cmds, output=False, **kwargs):
        """run git command"""
        cmds = [self._git_binary] + cmds  # prepend with git
        # output shall be captured
        LOG.debug('running git command: %s', ' '.join(cmds))
        if output:
            rval = subprocess.check_output(
                cmds,
                stderr=subprocess.STDOUT,
                **kwargs
            ).decode()
        # output does not matter, send it to /dev/null
        else:
            with open(os.devnull, 'w') as devnull:
                rval = subprocess.call(
                    cmds,
                    stdout=devnull,
                    stderr=devnull,
                    **kwargs
                )
        LOG.debug('git command return value: %s', rval)
        return rval

    @staticmethod
    def clean_workdir(workdir):
        """Remove workdir"""
        LOG.debug("Trying to remove workdir: %s", workdir)
        try:
            shutil.rmtree(workdir)
        except OSError:
            pass

    def verify_git(self, clone_url):
        """verify if git repository exists"""
        LOG.debug("Verifying if git repository %s is valid", clone_url)
        clone_url_hash = sha256(clone_url)
        workdir = os.path.join(
            self._tmpdir,
            'verify-git-%s' % clone_url_hash
        )
        if self.cmd(
                [
                    'clone',
                    clone_url,
                    'verify-git-%s' % clone_url_hash
                ],
                cwd=self._tmpdir
        ):
            LOG.debug('Could not verify git repository. Clone failed.')
            self.clean_workdir(workdir)
            raise RuntimeError('Could not verify git repository. Clone failed.')
        LOG.debug("%s is a valid git repository", clone_url)
        return workdir

    def verify_git_tag(self, tag, workdir):
        """verify if git tag exists in repository"""
        LOG.debug("Verifying if git tag %s is valid in %s", tag, workdir)
        try:
            gitout = self.cmd(['tag', '--list', tag], cwd=workdir, output=True)
        except subprocess.CalledProcessError:
            gitout = None

        if not gitout:
            LOG.debug('Could not verify git tag. Tag %s not found', tag)
            self.clean_workdir(workdir)
            raise RuntimeError(
                'Could not verify git tag. Tag %s not found' % tag
            )
        self.clean_workdir(workdir)
        LOG.debug("Git tag %s is valid", tag)

    def verify_git_branch(self, branch, workdir):
        """verify if git branch exists in repository"""
        LOG.debug("Verifying if git branch %s is valid in %s", branch, workdir)
        try:
            gitout = self.cmd(
                ['branch', '--list', '--all', "origin/%s" % branch],
                cwd=workdir, output=True
            )
        except subprocess.CalledProcessError:
            gitout = None

        if not gitout:
            LOG.debug(
                'Could not verify git branch. Branch %s not found', branch
            )
            self.clean_workdir(workdir)
            raise RuntimeError(
                'Could not verify git branch. Branch %s not found' % branch
            )
        self.clean_workdir(workdir)
        LOG.debug("Git branch %s is valid", branch)

    def verify_git_commit(self, commit, workdir):
        """verify if git commit exists in repository"""
        LOG.debug("Verifying if git commit %s is valid in %s", commit, workdir)
        try:
            gitout = self.cmd(
                ['cat-file', '-t', commit],
                cwd=workdir,
                output=True
            ).strip()
        except subprocess.CalledProcessError:
            gitout = None

        if gitout != "commit":
            LOG.debug(
                'Could not verify git commit. Commit %s not found', commit
            )
            self.clean_workdir(workdir)
            raise RuntimeError(
                'Could not verify git commit. Commit %s not found' % commit
            )
        self.clean_workdir(workdir)
        LOG.debug("Git commit %s is valid", commit)

    def get_module_info(self, clone_url, cache=True):
        """get latest tag from a git repository"""
        clone_url_hash = sha256(clone_url)
        version = 'unknown'
        date = None
        workdir = os.path.join(
            self._tmpdir,
            'get-module-info-%s' % clone_url_hash
        )

        local_info = {}
        if self._cache and cache:
            local_info = JsonCache(self._cache).read(clone_url_hash, 86400)
            LOG.debug('Get %s info from cache: %s', clone_url, local_info)

        if not local_info:
            try:
                if self.cmd(
                        [
                            'clone',
                            clone_url,
                            'get-module-info-%s' % clone_url_hash
                        ],
                        cwd=self._tmpdir
                ):
                    raise RuntimeError('clone failed')
                if self.cmd(
                        ['rev-list', '--tags', '--max-count=1'],
                        cwd=workdir
                ):
                    raise RuntimeError('rev-list failed')
                else:
                    # get sha1 for latest tag
                    cid = self.cmd(
                        ['rev-list', '--tags', '--max-count=1'],
                        output=True,
                        cwd=workdir
                    ).strip()
                    # get tag name from sha1
                    version = self.cmd(
                        ['describe', '--tags', cid],
                        output=True,
                        cwd=workdir
                    ).strip()
                    date = self.cmd(
                        ['show', '-s', '--format=%ci',
                         '%s^{commit}' % version],
                        output=True,
                        cwd=workdir
                    ).strip()
            except RuntimeError:
                pass
            finally:
                try:
                    shutil.rmtree(workdir)
                except OSError:
                    pass

            local_info = {
                'version': version,
            }
            if date:
                local_info['date'] = datetime.strptime(
                    date,
                    '%Y-%m-%d %H:%M:%S %z',
                ).strftime('%Y-%m-%d')

            if self._cache and cache:
                JsonCache(self._cache).write(clone_url_hash, local_info)
                LOG.debug('Written %s info to cache: %s', clone_url, local_info)

        return local_info

    def write_puppetfile(self, branch, puppetfile):
        """write Puppetfile and return diff"""
        LOG.debug("Update local control repository branch %s", branch)
        self.cmd(['checkout', branch], cwd=self._workdir)
        self.cmd(['pull', 'origin', branch], cwd=self._workdir)

        LOG.debug(
            "Trying to write puppetfile to %s",
            os.path.join(self._workdir, 'Puppetfile')
        )
        with open(os.path.join(self._workdir, 'Puppetfile'), 'w') as pfile:
            pfile.write('\n'.join(puppetfile))

        return self.cmd(['diff'], cwd=self._workdir, output=True)

    def commit_puppetfile(self, branch, message):
        """commit & push current version of Puppetfile"""
        self.cmd(['commit', '-m', message, 'Puppetfile'],
                 cwd=self._workdir)
        self.cmd(['push', 'origin', branch],
                 cwd=self._workdir)

    def revert_puppetfile(self):
        """revert uncommitted changes to Puppetfile"""
        self.cmd(['checkout', '--', 'Puppetfile'], cwd=self._workdir)

    def collect_puppetfiles(self, clone_url):
        """get Puppetfile from every branch of a r10k control repository"""
        puppetfiles = {}

        # clone git r10k control repository into a temporary directory
        self.cmd(['clone', clone_url, 'control'], cwd=self._tmpdir)
        self._workdir = '%s/control' % self._tmpdir

        # gather a list of local and remote branches
        branch_list = self.cmd(
            ['branch', '-a'],
            output=True,
            cwd=self._workdir
        ).split('\n')
        localbranches = []
        self.controlbranches = []
        for line in branch_list:
            stripped_line = line.strip(' *')
            if not stripped_line:
                continue
            if ' -> ' in stripped_line:
                continue
            if '/' in stripped_line:
                self.controlbranches.append(stripped_line.rsplit('/', 1)[1])
            else:
                localbranches.append(stripped_line)

        LOG.debug("Found controlbranches: %s", self.controlbranches)
        # go through list of branches, check them out and read Puppetfile into
        # memory
        for branch in self.controlbranches:
            LOG.debug("Process controlbranch: %s", branch)
            # if its a local branch, we can just check it out
            # and pull from origin to have it uptodate.
            if branch in localbranches:
                self.cmd(['checkout', branch], cwd=self._workdir)
                self.cmd(['pull', 'origin', branch], cwd=self._workdir)
            # if it is a remote branch, we first need to create a local branch
            # and can then pull from origin to have it uptodate.
            else:
                self.cmd(
                    ['checkout', '-b', branch, 'origin/%s' % branch],
                    cwd=self._workdir
                )

            # read puppetfile and store it memory
            LOG.debug(
                "Trying to read %s into memory",
                '%s/Puppetfile' % self._workdir
            )
            with open('%s/Puppetfile' % self._workdir, 'r') as puppetfile:
                puppetfiles[branch] = puppetfile.readlines()

        return puppetfiles



#!/usr/bin/env python3

"""manage a r10k-style control repository"""

import configparser
import fnmatch
import logging
import os
import re
from textwrap import TextWrapper
from .git import Git
from .forge import get_forge_module_info, verify_module_version
from .utils import cprint, fnlistmatch, query_yes_no, sha256
from .puppetmodules import PuppetModule, GitModule, ForgeModule
from .puppetmodules.versions import Forge, GitBranch, GitCommit, GitRef, GitTag

__author__ = "Andre Keller <andre.keller@vshn.ch>"
__copyright__ = "Copyright (c) 2015, VSHN AG, info@vshn.ch"
__license__ = 'BSD'
__version__ = '0.10.3'

LOG = logging.getLogger(__name__)


class Report:
    """control repository report"""

    def __init__(
            self,
            config,
            tmpdir,
            cache=None,
            args=None):
        self._branches = {}
        self._modules = {}

        self._cache = cache
        self._config = config
        self._args = args
        self._tmpdir = tmpdir

        self._git = Git(self._tmpdir, self._cache)
        self._process()

    @property
    def branches(self):
        """control repository branches"""
        return tuple(self._branches.keys())

    @property
    def modules(self):
        """all modules (from all control repository branches)"""
        return tuple(self._modules.keys())

    @staticmethod
    def _parse_mod_name(full_mod_name):
        """split full module name into module name and author"""
        mod_name = full_mod_name.rsplit('/', 1)
        if len(mod_name) > 1:
            return mod_name[1], mod_name[0]
        else:
            return mod_name[0], None

    @staticmethod
    def _extract_version_info(mod_line):
        """extract version information from module line"""
        version_info = {}
        for fragment in mod_line:
            clean = fragment.strip(' \'"')
            # if part does not start with a colon, it is a plain
            # version number (i.e. a forge module)
            if clean.startswith(':'):
                if clean.startswith(':git'):
                    version_info['url'] = clean.rsplit(
                        '>', 1
                    )[1].strip(' \'"')
                    version_info['type'] = 'git'
                elif clean.startswith(':commit'):
                    version_info['commit'] = clean.rsplit(
                        '>', 1
                    )[1].strip(' \'"')
                elif clean.startswith(':ref'):
                    version_info['ref'] = clean.rsplit(
                        '>', 1
                    )[1].strip(' \'"')
                elif clean.startswith(':tag'):
                    version_info['tag'] = clean.rsplit(
                        '>', 1
                    )[1].strip(' \'"')
                elif clean.startswith(':branch'):
                    version_info['branch'] = clean.rsplit(
                        '>', 1
                    )[1].strip(' \'"')
            else:
                version_info['version'] = clean
        LOG.debug('Extracted version info %s from %s', version_info, mod_line)
        return version_info

    @staticmethod
    def _collapse_puppetfile_lines(puppetfile):
        """drop non-mod lines and collapse mod lines into a single line"""
        lines = []
        re_comment = re.compile(r'^(\s+)?#')
        re_mod = re.compile(r'^(\s+)?mod')
        line_buffer = None
        LOG.debug('Collapsing Puppetfile lines')
        for line in puppetfile:
            stripped_line = line.strip()
            # skip emtpy lines
            if not stripped_line:
                LOG.debug("Skipped empty line")
                continue
            # skip comments
            if re_comment.match(stripped_line):
                LOG.debug("Skipped comment: %s", stripped_line)
                continue
            # we are processing a multiline mod block
            if line_buffer:
                # append current line to buffer
                line_buffer = line_buffer + stripped_line
                LOG.debug('Added fragment to buffer: %s', stripped_line)
                # if line does not end with a comma,
                # means we are done processing this block
                if not stripped_line.endswith(','):
                    lines.append(line_buffer)
                    LOG.debug(
                        'Finished multiline block, added mod line: %s',
                        line_buffer
                    )
                    line_buffer = None
                    continue
            # unless we are processing a multiline block
            # only mod lines are interesting
            if re_mod.match(stripped_line):
                if stripped_line.endswith(','):
                    LOG.debug('Processing multiline block')
                    line_buffer = stripped_line
                    LOG.debug('Added fragment to buffer: %s', stripped_line)
                else:
                    lines.append(stripped_line)
                    LOG.debug('Added mod line: %s', stripped_line)
        return lines

    @staticmethod
    def _git_module_version(module_data):
        """returns git version object, based on module_data"""
        if 'branch' in module_data:
            return GitBranch(module_data['branch'])
        elif 'commit' in module_data:
            return GitCommit(module_data['commit'])
        elif 'ref' in module_data:
            return GitRef(module_data['ref'])
        elif 'tag' in module_data:
            return GitTag(module_data['tag'])
        return None


    def _module_object(self, name, module_data):
        """Returns a PuppetModule object for module_data"""
        if module_data['type'] == 'forge':
            forge_module = ForgeModule(name, module_data['author'])
            if 'version' in module_data:
                forge_module.version = Forge(module_data['version'])
            if self._args.version_check:
                forge_module.info = get_forge_module_info(
                    forge_module.forgename, self._cache
                )
            return forge_module
        else:
            git_module = GitModule(name, module_data['url'])
            if self._args.version_check:
                git_module.info = self._git.get_module_info(git_module.url)
                git_module.version = self._git_module_version(module_data)
            return git_module

    @staticmethod
    def forge_module_from_args(args):
        """Returns a ForgeModule object based on cli args"""
        module = ForgeModule(
            args.module.split('/')[1],
            args.module.split('/')[0]
        )
        if args.version:
            if args.version == 'LATEST_FORGE_VERSION':
                try:
                    forge_info = get_forge_module_info(args.module)
                    module.version = Forge(forge_info['version'])
                except (KeyError, TypeError) as exc:
                    raise RuntimeError(
                        "could not gather latest %s version" % args.module
                    ) from exc
            else:
                verify_module_version(args.module, args.version)
                module.version = Forge(args.version)
        else:
            try:
                forge_info = get_forge_module_info(args.module)
                _ = forge_info['version']
            except (KeyError, TypeError) as exc:
                raise RuntimeError(
                    "could find %s on forge" % args.module
                ) from exc
        return module

    def git_module_from_args(self, args):
        """Returns a GitModule object based on cli args"""
        module = GitModule(args.module.split("/")[-1], args.git)
        workdir = self._git.verify_git(args.git)
        if args.tag:
            if args.tag == 'LATEST_GIT_TAG':
                try:
                    git_info = self._git.get_module_info(args.git, False)
                    module.version = GitTag(git_info['version'])
                except (KeyError, TypeError) as exc:
                    raise RuntimeError(
                        "could not gather latest %s version" % args.module
                    ) from exc
            else:
                self._git.verify_git_tag(args.tag, workdir)
                module.version = GitTag(args.tag)
        elif args.commit:
            self._git.verify_git_commit(args.commit, workdir)
            module.version = GitCommit(args.commit)
        elif args.branch:
            self._git.verify_git_branch(args.branch, workdir)
            module.version = GitBranch(args.branch)
        else:
            self._git.clean_workdir(workdir)
        return module

    def module_from_args(self, args):
        """Returns a puppet module object based on cli args"""
        if not args.module:
            return None

        if args.remove:
            return PuppetModule(args.module.split("/")[-1])

        if args.git:
            return self.git_module_from_args(args)
        else:
            return self.forge_module_from_args(args)

    def _process(self):
        """process puppetfiles, extract module information"""
        puppetfiles = self._git.collect_puppetfiles(
            self._config.get('crmngr', 'repository')
        )

        for branch, puppetfile in puppetfiles.items():
            LOG.debug('Processing Puppetfile for branch: %s', branch)
            self._branches.setdefault(branch, [])

            # gather branch information (modules for each branch)
            mod_lines = self._collapse_puppetfile_lines(puppetfile)
            for mod_line in mod_lines:
                # split module line into comma-separated parts (starting after
                # 'mod ')
                line_parts = mod_line[4:].split(',')

                # if module_name contains a /, split it into
                # mod_name / mod_author
                mod_name, mod_author = self._parse_mod_name(
                    line_parts[0].strip(' \'"')
                )

                module_data = {}

                if mod_author:
                    module_data['author'] = mod_author
                    module_data['type'] = 'forge'
                # parse additional parts of mod line
                module_data.update(
                    self._extract_version_info(line_parts[1:])
                )

                # create module object
                module = self._module_object(mod_name, module_data)
                # add branch to modules used branch info
                module.add_branch(branch)

                # add module to module list, if not yet there
                self._modules.setdefault(module.name, {})

                # if current module version is in reports module list,
                # add branch to module versions used branch info
                if module in self._modules[module.name]:
                    self._modules[module.name][module].add_branch(branch)
                    LOG.debug(
                        "Added branch %s to %s's used branches",
                        branch, module
                    )
                else:
                    # add module version to reports module list.
                    self._modules[module.name].setdefault(module, module)
                    LOG.debug("Added module %s to reports module list", module)

                # also add module to reports branch list.
                self._branches[branch].append(module)
                LOG.debug(
                    "Added module %s to branch %s module list",
                    module, branch
                )

    def _get_module_categories(self):
        """Get categories from configuration file"""
        try:
            categories = [i.strip() for i in self._config.get(
                'crmngr', 'categories'
            ).split(',')]
            categories.append("uncategorized")
        except (configparser.NoOptionError, configparser.NoSectionError):
            categories = ["uncategorized"]
        LOG.debug("Categories defined in configuration file: %s", categories)
        return categories

    def _get_module_category(self, module, categories):
        """Get category for specific module (from configuration file)"""

        for category in categories:
            try:
                self._config.get(category, module)
                return category
            except (configparser.NoOptionError, configparser.NoSectionError):
                pass
        return "uncategorized"

    def prepare_puppetfiles(self, mode, branches, update_module=None):
        """return generated puppetfile for a specific branch"""
        puppetfiles = {}

        for branch in sorted(self._branches.keys()):
            LOG.debug("Prepare puppetfile for branch: %s", branch)
            if not fnlistmatch(branch, branches):
                LOG.debug(
                    "Skipped branch %s, not in requested branches list: %s",
                    branch,
                    branches
                )
                continue
            lines = [
                "forge 'http://forge.puppetlabs.com'",
                '',
            ]

            categories = self._get_module_categories()

            if mode == 'add':
                for module in self._branches[branch]:
                    if module.name == update_module.name:
                        break
                else:
                    self._branches[branch].append(update_module)
                    LOG.debug("Added module %s", update_module)

            module_lines = {}
            for module in sorted(self._branches[branch]):
                LOG.debug("Prepare module %s", module)
                if isinstance(update_module, PuppetModule):
                    if module.name == update_module.name:
                        if mode == 'remove':
                            LOG.debug("Remove module %s")
                            continue
                        module = update_module
                        LOG.debug(
                            "Ensure module version %s for module %s",
                            update_module, module.name
                        )
                module_category = self._get_module_category(
                    module.name, categories
                )
                module_lines.setdefault(module_category, [])
                module_lines[module_category] += module.puppetfile
                LOG.debug(
                    "Added module %s to puppetfile category %s",
                    module, module_category
                )

            for category in [i for i in categories if i in module_lines]:
                if len(categories) > 1:
                    lines.append("# %s modules ###" % category.upper())
                lines += module_lines[category]
                lines.append('')

            puppetfiles[branch] = lines

        return puppetfiles

    def update_puppetfiles(self, puppetfiles, commit_message):
        """update puppetfile"""
        for branch, puppetfile in sorted(puppetfiles.items()):
            LOG.debug("Write puppetfile for branch %s", branch)
            diff = self._git.write_puppetfile(branch, puppetfile)
            if diff:
                LOG.debug(
                    "Puppetfile differences for branch %s: %s", branch, diff
                )
                if not self._args.debug:
                    if not (self._args.diffonly or self._args.noninteractive):
                        if self._config.has_option('crmngr',
                                                   'console_clear_command'):
                            os.system(self._config.get('crmngr',
                                                       'console_clear_command'))
                if self._args.diffonly:
                    cprint.white_bold("Diff for branch %s" % branch)
                    cprint.diff(diff)
                    self._git.revert_puppetfile()
                elif self._args.noninteractive:
                    print('Updating Puppetfile in branch %s' % branch)
                    self._git.commit_puppetfile(branch, commit_message)
                else:
                    cprint.white_bold("Diff for branch %s" % branch)
                    cprint.diff(diff)
                    if query_yes_no("Update (commit and push) puppetfile for "
                                    "branch %s" % branch):
                        self._git.commit_puppetfile(branch, commit_message)
                    else:
                        self._git.revert_puppetfile()
            else:
                LOG.debug("Puppetfile for branch %s unchanged", branch)

    def print_module_report(self,
                            report_modules=None,
                            report_branches=None,
                            report_unused_branches=False):
        """print report aggregated by module"""

        requested_branches = []
        for module in sorted(self.modules):
            # check if we need to report this module at all...
            if not fnlistmatch(module, report_modules):
                LOG.debug(
                    'Module %s does not match requested modules pattern (%s). '
                    'Skipping',
                    module, report_modules
                )
                continue

            LOG.debug("Process module: %s", module)
            match = False
            for _, module_details in self._modules[module].items():
                if report_branches:
                    for requested_branch in report_branches:
                        for branch in module_details.branches:
                            if fnmatch.fnmatch(branch, requested_branch):
                                requested_branches.append(branch)
                                match = True
                else:
                    requested_branches = list(self.branches)
                    match = True

            LOG.debug(
                'Requested branches: %s (from pattern %s)',
                requested_branches, report_branches
            )

            if not match:
                LOG.debug(
                    'Module %s not in any of the requested branches. Skipping'
                )
                continue

            cprint.white_bold('Module: %s' % module)

            module_branches = []
            for _, module_details in sorted(
                    self._modules[module].items(), reverse=True):
                LOG.debug("Process module version %s", module_details)
                if not [i for i in module_details.branches
                        if i in set(requested_branches)]:
                    LOG.debug(
                        "Module version %s not in any of the requested "
                        "branches (%s). Skipping",
                        module_details,
                        requested_branches
                    )
                    continue
                module_branches += module_details.branches
                module_details.print_version_information()
                self._print_used_branches(
                    sorted(module_details.branches),
                    set(requested_branches)
                )
                print()
            if report_unused_branches:
                self._print_unused_branches(
                    module,
                    sorted(list(set(requested_branches) - set(module_branches)))
                )
                print()

    @staticmethod
    def _print_unused_branches(module, branches):
        """print list of branches that do not use a specific module"""
        if branches:
            cprint.white('Branches not using %s module:' % module, lpad=2)
            branch_wrapper = TextWrapper(
                initial_indent=' ' * 4,
                subsequent_indent=' ' * 4
            )
            for line in branch_wrapper.wrap(", ".join(branches)):
                cprint.yellow(line)

    @staticmethod
    def _print_used_branches(branches, requested_branches):
        """print list of branches used by a specific module"""
        if requested_branches:
            local_branches = [i for i in branches if i in requested_branches]
        else:
            local_branches = branches
        if local_branches:
            cprint.white('Used by:', lpad=4, rpad=4, end='')
            branch_wrapper = TextWrapper(
                subsequent_indent=' ' * 16
            )
            for line in branch_wrapper.wrap(", ".join(local_branches)):
                cprint.cyan(line)



#!/usr/bin/env python3

"""collection of functions used by crmngrlib"""

import fnmatch
import hashlib
import sys


def sha256(string):
    """generate sha256 sum"""
    return hashlib.sha256(string.encode('utf-8')).hexdigest()


def fnlistmatch(name, patterns, noneistrue=True):
    """fnmatch a list of patterns"""
    if not patterns:
        if noneistrue:
            return True
        return False

    for pattern in patterns:
        if fnmatch.fnmatch(name, pattern):
            return True
    return False


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer."""
    valid = {"yes": True, "y": True, "ye": True, "j": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")



#!/usr/bin/env python3

"""puppet forge related functions for crmngrlib"""

import logging
import json

import urllib.request
from urllib.error import URLError

from datetime import datetime

from .cache import JsonCache
from .utils import sha256

LOG = logging.getLogger(__name__)


def get_forge_module_info(module, cache=None):
    """get current release from forge api"""

    normalized_module = '%s-%s' % (
        module.split('/', 1)[0], module.split('/', 1)[1]
    )

    module_url = '%s/%s' % (
        'https://forgeapi.puppetlabs.com:443/v3/modules',
        normalized_module
    )

    local_info = {}
    if cache:
        local_info = JsonCache(cache).read(sha256(normalized_module), 86400)
        LOG.debug('Get %s info from cache: %s', normalized_module, local_info)

    if not local_info:
        try:
            api_request = urllib.request.Request(module_url)
            api_response = urllib.request.urlopen(
                api_request
            ).read().decode('utf-8')

            api_info = json.loads(api_response)
            LOG.debug('Retrieved module info from %s', module_url)

            local_info = {
                'version': api_info['current_release']['version'],
                'date': datetime.strptime(
                    api_info['current_release']['updated_at'],
                    '%Y-%m-%d %H:%M:%S %z',
                ).strftime('%Y-%m-%d')
            }
        except (AttributeError, KeyError, URLError, ValueError):
            return dict()

        if cache:
            JsonCache(cache).write(sha256(normalized_module), local_info)
            LOG.debug(
                'Written %s info to cache: %s',
                normalized_module,
                local_info
            )

    return local_info


def verify_module_version(module, version):
    """get current release from forge api"""

    normalized_module = '%s-%s' % (
        module.split('/', 1)[0], module.split('/', 1)[1]
    )

    module_url = '%s/%s' % (
        'https://forgeapi.puppetlabs.com:443/v3/modules',
        normalized_module
    )

    try:
        api_request = urllib.request.Request(module_url)
        api_response = urllib.request.urlopen(
            api_request
        ).read().decode('utf-8')

        api_info = json.loads(api_response)
        LOG.debug('Retrieved module info from %s', module_url)

        for release in api_info['releases']:
            if version == release['version']:
                LOG.debug('Found version %s in release versions', version)
                break
        else:
            raise RuntimeError('Version %s not found' % version)

    except (AttributeError, KeyError, URLError, ValueError) as exc:
        raise RuntimeError('Failed to determine module version') from exc



#!/usr/bin/env python3

"""function for colored terminal output"""

TERM_BLUE = '\033[0;34m'
TERM_BLUE_BOLD = '\033[1;34m'
TERM_CYAN = '\033[0;36m'
TERM_CYAN_BOLD = '\033[1;36m'
TERM_GREEN = '\033[0;32m'
TERM_GREEN_BOLD = '\033[1;32m'
TERM_MAGENTA = '\033[0;35m'
TERM_MAGENTA_BOLD = '\033[1;35m'
TERM_RED = '\033[0;31m'
TERM_RED_BOLD = '\033[1;31m'
TERM_WHITE = '\033[0;37m'
TERM_WHITE_BOLD = '\033[1;37m'
TERM_YELLOW = '\033[0;33m'
TERM_YELLOW_BOLD = '\033[1;33m'

TERM_NONE = '\033[0;m'


def _cprint(color, text, **kwargs):
    """helper function to print colored text"""
    prefix = kwargs.pop('prefix', '')
    suffix = kwargs.pop('suffix', '')
    lpad = kwargs.pop('lpad', 0)
    rpad = kwargs.pop('rpad', 0)
    sep = kwargs.pop('sep', '')

    print(
        ' '*lpad,
        prefix,
        color,
        text,
        suffix,
        ' '*rpad,
        TERM_NONE,
        sep=sep,
        **kwargs
    )


def blue(text, **kwargs):
    """print blue text"""
    _cprint(TERM_BLUE, text, **kwargs)


def blue_bold(text, **kwargs):
    """print bold blue text"""
    _cprint(TERM_BLUE_BOLD, text, **kwargs)


def cyan(text, **kwargs):
    """print cyan text"""
    _cprint(TERM_CYAN, text, **kwargs)


def cyan_bold(text, **kwargs):
    """print bold cyan text"""
    _cprint(TERM_CYAN_BOLD, text, **kwargs)


def green(text, **kwargs):
    """print green text"""
    _cprint(TERM_GREEN, text, **kwargs)


def green_bold(text, **kwargs):
    """print bold green text"""
    _cprint(TERM_GREEN_BOLD, text, **kwargs)


def magenta(text, **kwargs):
    """print magenta text"""
    _cprint(TERM_MAGENTA, text, **kwargs)


def magenta_bold(text, **kwargs):
    """print bold magenta text"""
    _cprint(TERM_MAGENTA_BOLD, text, **kwargs)


def red(text, **kwargs):
    """print red text"""
    _cprint(TERM_RED, text, **kwargs)


def red_bold(text, **kwargs):
    """print bold red text"""
    _cprint(TERM_RED_BOLD, text, **kwargs)


def white(text, **kwargs):
    """print white text"""
    _cprint(TERM_WHITE, text, **kwargs)


def white_bold(text, **kwargs):
    """print bold white text"""
    _cprint(TERM_WHITE_BOLD, text, **kwargs)


def yellow(text, **kwargs):
    """print yellow text"""
    _cprint(TERM_YELLOW, text, **kwargs)


def yellow_bold(text, **kwargs):
    """print bold yellow text"""
    _cprint(TERM_YELLOW_BOLD, text, **kwargs)


def diff(text):
    """print colorized diff"""
    for line in text.split('\n'):
        if line.startswith('-'):
            red(line)
        elif line.startswith('+'):
            green(line)
        else:
            white(line)



"""
crmngr setup module.
"""

from setuptools import setup, find_packages

import crmngrlib

setup(
    name='crmngr',
    version=crmngrlib.__version__,
    description='manage a r10k-style control repository',
    url='https://github.com/vshn/crmngr',
    author='Andre Keller',
    author_email='andre.keller@vshn.ch',
    # BSD 3-Clause License:
    # - http://opensource.org/licenses/BSD-3-Clause
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: System :: Systems Administration',
    ],

    packages=[
        'crmngrlib',
        'crmngrlib.puppetmodules',
        'crmngrlib.puppetmodules.versions',
        'crmngrlib.utils',
    ],

    scripts=['crmngr',],

)



#!/usr/bin/env python3

"""cache related functions for crmngrlib"""

import json
import logging
import os
import time

LOG = logging.getLogger(__name__)


class CacheError(Exception):
    """custom exception for cache related errors"""
    pass


class JsonCache:
    """json file based cache"""

    def __init__(self, directory, fail_silently=True):
        """constructor, takes directory as argument"""
        LOG.debug("Initialize JsonCache in %s", directory)
        self.directory = directory
        self.fail_silently = fail_silently

    def read(self, key, ttl=86400):
        """read json dict from file"""
        try:
            LOG.debug("Try to read %s from cache", key)
            with open(os.path.join(self.directory, key)) as cache_fd:
                cache = json.load(cache_fd)
                LOG.debug("Got %s from cache", cache)
            if cache.get('updated', 0) + ttl >= int(time.time()):
                LOG.debug("Cache entry is valid, return it")
                return cache
            LOG.debug("Cache expired, returning empty response")
            return dict()
        except (AttributeError, KeyError, OSError, ValueError) as exc:
            if not self.fail_silently:
                raise CacheError('could not read from cache') from exc
            LOG.debug(
                "Cache lookup for %s failed. Fail silently.", key
            )

    def write(self, key, jsondict):
        """write json dict to file"""
        try:
            LOG.debug("Try to write %s to cache using key %s", jsondict, key)
            with open(os.path.join(self.directory, key), 'w') as cache_fd:
                localdict = jsondict.copy()
                localdict.update(
                    {
                        'updated': int(time.time()),
                    }
                )
                LOG.debug('Set updated to current time')
                json.dump(localdict, cache_fd)
                LOG.debug('Wrote %s to cache using key %s', localdict, key)
        except (AttributeError, KeyError, OSError, ValueError) as exc:
            if not self.fail_silently:
                raise CacheError('could not write to cache') from exc
            LOG.debug('Failed to write to cache. Fail silently.')
